;; Generated package description from isl.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "isl" "1.6.0.20260414.62054" "Simple incremental search in current-buffer" '((emacs "29.1")) :commit "a998ee1c9a1b682f6fbc5eced4c8e22605b65eef" :authors '(("Thierry Volpiatto" . "thievol@posteo.net")) :maintainer '("Thierry Volpiatto" . "thievol@posteo.net") :keywords '("matching" "searching") :url "https://github.com/thierryvolpiatto/isearch-light")
